from ._dcca import DCCA
from ._dcca_barlow_twins import BarlowTwins
from ._dcca_eigengame import DCCA_EigenGame
from ._dcca_gha import DCCA_GHA
from ._dcca_noi import DCCA_NOI
from ._dcca_sdl import DCCA_SDL
from ._dtcca import DTCCA
